PROGRAM Backpropagation;

CONST
      n_max      = 10;   { maximale Anzahl von Neuronen pro Schicht }
      m_max      = 5;    { maximale Anzahl verschiedener Musterpaare }

TYPE
      t_synapsen   = ARRAY [1..n_max, 1..n_max] OF real;
      t_musterpaar = RECORD
                       e : ARRAY [1..n_max] OF real;  { Eingabewerte }
                       s : ARRAY [1..n_max] OF real; {Ausgabesollwerte }
                     END;

VAR                                        { global }
         anzahl_E, anzahl_I,
         anzahl_A             : 1..n_max;
         anzahl_m             : 1..m_max;
         wai, wie             : t_synapsen;
         muster               : ARRAY [1..m_max] OF t_musterpaar;

PROCEDURE initialisierung;
      BEGIN
      { Eingabe von anzahl_E :      Anzahl der Eingabeneuronen }
      { Eingabe von anzahl_A :      Anzahl der Ausgabeneuronen }
      { Eingabe von anzahl_I :      Anzahl der inneren Neuronen }
      { Belegung von wie mit :      Synapsengewichte zwischen den inneren }
      {                             zuflligen Werten Neuronen und den }
      {                             Eingabeneuronen }
      { Belegung von wai mit    :   Synapsengewichte zwischen den Ausgabe }
      {                             zuflligen Werten neuronen und den }
      {                             inneren Neuronen }
      { Eingabe von anzahl_M :      Anzahl der verschiedenen Musterpaare }
      { Eingabe der Musterpaare}
END;

FUNCTION sigma (x : real) : real;  { Normierungsfunktion }
BEGIN
 sigma := 1 / (1 + exp (-x))
END; { sigma }


FUNCTION aktivitaet_I (i,m : integer) : real;
{ Aktivitt des inneren Neurons i bei Muster m }
VAR      x : real;
         e : integer;
BEGIN
      x := 0;
      FOR e:=1 TO anzahl_e DO x := x + wie [i,e] * muster [m].e [e];
      aktivitaet_I := sigma (x)
END; { aktivitaet_I }


FUNCTION aktivitaet_A (a,m : integer) : real;
{ Aktivitt des Ausgabeneurons a bei Musterpaar m }
VAR      x : real;
         i : integer;
BEGIN
      x := 0;
      FOR i:=1 TO anzahl_i DO x := x+ wai [a,i] * aktivitaet_I (i,m);
      aktivitaet_A := sigma (x)
END; { aktivitaet_A }


FUNCTION abweichung (a,m : integer) : real;
{ Ausgabefehler des Ausgabeneurons a bei Muster m }
BEGIN
      abweichung := muster [m].s [a] - aktivitaet_A (a,m)
END; { abweichung }

FUNCTION korrektur_AI (a,i,m : integer) : real;
{ Korrektur der Synapsengewichte zwischen }
{ inneren Neuronen und Ausgabeneuronen  }
BEGIN
      korrektur_AI := abweichung (a,m) * aktivitaet_I (i,m) *
                      aktivitaet_A (a,m) * ( 1 - aktivitaet_A (a,m))
END; { korrektur_AI }


FUNCTION korrektur_IE (i,e,m : integer) : real;
{ Korrektur der Synapsengewichte zwischen }
{ Eingabeneuronen und inneren Neuronen  }

VAR  a: integer;
     summe: real;
BEGIN
      summe := 0;
      FOR a := 1 TO anzahl_A DO
        summe := summe + abweichung (a,m) * muster [m].e [e] *
                 aktivitaet_A (a,m) * (1 - aktivitaet_A (a,m)) *
                 wai [a,i] * aktivitaet_I (i,m) * (1 - aktivitaet_I (i,m));
      korrektur_IE := summe
END; { korrektur_IE }

PROCEDURE korrektur (alpha : real; m : integer);
VAR      a,i,e       : integer;
         kai, kie    : t_synapsen;

BEGIN
      FOR i:=1 TO anzahl_I DO
         BEGIN
               FOR a:=1 TO anzahl_A DO kai [a,i] := korrektur_AI (a,i,m);
               FOR e:=1 TO anzahl_E DO kie [i,e] := korrektur_IE (i,e,m);
         END; { FOR }

      FOR i:=1 TO anzahl_I DO
         BEGIN
           FOR a:=1 TO anzahl_A DO
                   wai [a,i] := wai [a,i] + alpha * kai [a,i];
           FOR e:=1 TO anzahl_E DO
                   wai [a,i] := wie [i,e] + alpha * kie [i,e];
         END; { FOR }
END; { Korrektur }

PROCEDURE lernen;
VAR      lernschritte,l,m   : integer;
         alpha              : real;

BEGIN
 { Eingabe der Anzahl der Lernschritte und der Lernschrittgre }
      FOR l:=1 TO lernschritte DO
         BEGIN
               { Whle per Zufall ein Muster aus }
               korrektur (alpha,m)
      END; { FOR }
END; { lernen }


PROCEDURE ausgabe;
BEGIN
 { Ausgabe der Neuronenaktivitten }
 { Die Prozedur ist an den verwendeten Compiler anzupassen }
END;


BEGIN { Hauptprogramm }
      initialisierung;
      lernen;
      ausgabe
END.
